# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# fmt: off
from .....schema_classes import GenericAspectClass
from .....schema_classes import MetadataChangeEventClass
from .....schema_classes import MetadataChangeProposalClass
from .....schema_classes import SystemMetadataClass


GenericAspect = GenericAspectClass
MetadataChangeEvent = MetadataChangeEventClass
MetadataChangeProposal = MetadataChangeProposalClass
SystemMetadata = SystemMetadataClass

# fmt: on
