{
  "type": "record",
  "name": "GlossaryTermKey",
  "namespace": "com.linkedin.pegasus2avro.metadata.key",
  "doc": "Key for a GlossaryTerm",
  "fields": [
    {
      "name": "name",
      "type": "string",
      "Searchable": {
        "enableAutocomplete": true,
        "fieldType": "TEXT_PARTIAL"
      }
    }
  ],
  "Aspect": {
    "name": "glossaryTermKey"
  }
}